unit rtcPhpProvider;

interface

uses
  Windows, SysUtils, Classes, 

  Forms, // D4

  rtcLog,
  rtcInfo, rtcConn,
  rtcDataSrv,
  rtcPHP,

  rtcFileProvider;

type
  TPHP_Provider = class(TDataModule)
    PhpProvider: TRtcDataProvider;
    ServerLink: TRtcDataServerLink;

    procedure PhpProviderCheckRequest(Sender: TRtcConnection);
    procedure PhpProviderDisconnect(Sender: TRtcConnection);
    procedure PhpProviderListenStart(Sender: TRtcConnection);
    procedure PhpProviderListenStop(Sender: TRtcConnection);
    procedure PhpProviderDataReceived(Sender: TRtcConnection);

    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);

  private
    { Private declarations }
    FReady:boolean;

    FDllFolder,
    FIniFolder:string;

    ExtList:TStringList;

  public
    { Public declarations }
    procedure ClearExts;
    procedure AddExts(a:string);

    property DllFolder:string read FDllFolder write FDllFolder;
    property IniFolder:string read FIniFolder write FIniFolder;
  end;

function GetPHPProvider:TPHP_Provider;

implementation

{$R *.dfm}

var
  PHP_Provider: TPHP_Provider;

function GetPHPProvider:TPHP_Provider;
  begin
  if not assigned(PHP_Provider) then
    PHP_Provider:=TPHP_Provider.Create(nil);
  Result:=PHP_Provider;
  end;

procedure TPHP_Provider.ClearExts;
  begin
  ExtList.Clear;
  end;

procedure TPHP_Provider.AddExts(a: string);
  var
    elist,ext:string;
  begin
  elist:=Trim(a);
  while elist<>'' do
    begin
    if Pos(',',elist)>0 then
      begin
      ext:=UpperCase(Trim(Copy(elist,1,Pos(',',elist)-1)));
      if Copy(ext,1,1)<>'.' then ext:='.'+ext;
      Delete(elist,1,Pos(',',elist));
      elist:=Trim(elist);
      end
    else
      begin
      ext:=UpperCase(elist);
      if Copy(ext,1,1)<>'.' then ext:='.'+ext;
      elist:='';
      end;
    ExtList.Add(ext);
    end;
  end;

procedure TPHP_Provider.PhpProviderListenStart(Sender: TRtcConnection);
  begin
  try
    StartupPHP(DllFolder, IniFolder);
    FReady:=True;
    XLog('PHP Ready.');
  except
    on E:Exception do
      XLog('Error starting PHP Library: '+E.Message);
    end;
  end;

procedure TPHP_Provider.PhpProviderListenStop(Sender: TRtcConnection);
  begin
  if FReady then
    begin
    try
      ShutDownPHP;
    except
      on E:Exception do
        XLog('Error shuting down PHP Library: '+E.Message);
      end;
    FReady:=False;
    end;
  end;

procedure TPHP_Provider.PhpProviderCheckRequest(Sender: TRtcConnection);

  procedure CheckPHPFile;
    var
      a:integer;
      MyFileName:string;
      DocRoot:string;
    begin
    with TRtcDataServer(Sender) do
      begin
      { Check the request header when it started and accept it immediatelly,
        even if we won't be responding before the request body is received. }
      DocRoot:=GetDocRoot(Request.Host);
      if DocRoot<>'' then
        begin
        MyFileName:=RepairWebFileName(Request.FileName,DocRoot);
        if MyFileName<>'' then
          begin
          for a:=0 to ExtList.Count-1 do
            begin
            if UpperCase(Copy(MyFileName,
                              length(MyFileName)-length(ExtList.Strings[a])+1,
                              length(ExtList.Strings[a]))) = ExtList.Strings[a] then
              begin
              Accept; // Accept the request (we will be responding)
              Request.FileName:=MyFileName;
              Break;
              end;
            end;
          end;
        end;
      end;
    end;

  begin
  with TRtcDataServer(Sender).Request do
    if (Method='GET') or (Method='POST') or (Method='HEAD') then
      CheckPHPFile;
  end;

procedure TPHP_Provider.PhpProviderDataReceived(Sender: TRtcConnection);

  procedure SendPHP(SendContent:boolean=True);
    var
      s,s2:string;
      TC:Cardinal;

    function Parse_PHP(Sender:TRtcDataServer; const filename:string):string;
      var
        PHP:TRtcPHP;
        s:string;
        a,TC:longword;
      begin
      PHP:=TRtcPHP.Create;
      try
        PHP.LocalAddr:=Sender.LocalAddr;
        PHP.LocalPort:=Sender.LocalPort;
        PHP.PeerAddr:=Sender.PeerAddr;
        PHP.PeerPort:=Sender.PeerPort;

        PHP.Request:=Sender.Request;
        PHP.RequestBody:=Sender.Read;

        PHP.DocumentRoot:=GetDocRoot(PHP.Request.Host);

        PHP.FileName:=filename;

        s2:='';
        try
          s:='';
          with Sender.Request.Query do if Text<>'' then
            begin
            s:='QUERY >> (Delimiter='+Delimiter+')'#13#10;
            for a:=0 to ItemCount-1 do
              s:=s+IntToStr(a+1)+'. "'+ItemName[a]+'" = <'+ItemValue[a]+'>'+#13#10;
            end;
        except
          on E:Exception do
            begin
            XLog('Request.Query error: '+E.Message,'PHP');
            s:='';
            end;
          end;
        s2:=s2+s;

        try
          s:='';
          with Sender.Request.Cookie do if Text<>'' then
            begin
            s:='COOKIE >> (Delimiter='+Delimiter+')'#13#10;
            for a:=0 to ItemCount-1 do
              s:=s+IntToStr(a+1)+'. "'+ItemName[a]+'" = <'+ItemValue[a]+'>'+#13#10;
            end;
        except
          on E:Exception do
            begin
            XLog('Request.Cookie error: '+E.Message,'PHP');
            s:='';
            end;
          end;
        s2:=s2+s;

        if PHP.RequestBody<>'' then
          begin
          try
            s:='';
            with Sender.Request.Params do
              begin
              Text:=PHP.RequestBody;
              s:=#13#10'PARAMS >> (Delimiter='+Delimiter+')'#13#10;
              for a:=0 to ItemCount-1 do
                s:=s+IntToStr(a+1)+'. "'+ItemName[a]+'" = <'+ItemValue[a]+'>'+#13#10;
              s:='<'+PHP.RequestBody+'>'+s;
              end;
          except
            on E:Exception do
              begin
              XLog('Request.Params error: '+E.Message,'PHP');
              s:='<'+PHP.RequestBody+'>';
              end;
            end;
          s2:=s2+s;
          end;

        XLog('Exec '+Sender.Request.Host+' > '+Sender.Request.URI+#13#10+
             Sender.Request.HeaderText+
             s2,
             'PHP');

        TC:=GetTickCount;
        Result:=PHP.Execute;
        TC:=GetTickCount-TC;

        Sender.Response.HeaderText:=PHP.ResultHeader;

        s2:='';
        try
          s:='';
          with Sender.Response.Cookie do if Text<>'' then
            begin
            s:='COOKIE >> (Delimiter='+Delimiter+')'#13#10;
            for a:=0 to ItemCount-1 do
              s:=s+IntToStr(a+1)+'. "'+ItemName[a]+'" = <'+ItemValue[a]+'>'+#13#10;
            end;
        except
          on E:Exception do
            begin
            XLog('Response.Cookie error: '+E.Message,'PHP');
            s:='';
            end;
          end;
        s2:=s2+s;

        XLog('Done '+Sender.Request.Host+' > '+Sender.Request.URI+' > '+IntToStr(TC)+' ms'#13#10+
             Sender.Response.HeaderText+
             s2,
             'PHP');

      finally
        PHP.Free;
        end;
      end;

    begin
    with TRtcDataServer(Sender) do
      begin
      TC:=GetTickCount;
      try
        Processing; // Signal that we will be processing a request (could take a while)
        if FReady then
          s:=Parse_PHP(TRtcDataServer(Sender),
                       StringReplace(GetDocRoot(Request.Host)+Request.FileName,'/','\',[rfReplaceAll]))
        else
          s:='ERROR! PHP engine not initialized.';
      except
        on E:Exception do
          s:=E.Message;
        end;
      TC:=GetTickCount-TC;
      XLog('SEND '+PeerAddr+' > '+Request.Host+
           ' "'+Request.Method+' '+Request.URI+'"'+
           ' '+IntToStr(length(s))+
           ' REF "'+Request.Referer+'"'+
           ' AGENT "'+Request.Agent+'" > '+IntToStr(TC)+' ms');

      if Response.ContentLength=0 then
        Response.ContentLength:=length(s);

      Response.SendContent:=SendContent;
      WriteHeader;

      if SendContent and (length(s)>0) then
        Write(s);
      end;
    end;

  begin
  // start processing when complete request body was received
  with TRtcDataServer(Sender) do
    if Request.Complete then
      SendPHP(Request.Method<>'HEAD');
  end;

procedure TPHP_Provider.PhpProviderDisconnect(Sender: TRtcConnection);
  begin
  with TRtcDataServer(Sender) do
    begin
    if Request.DataSize>Request.DataIn then
      begin
      // did not receive a complete request
      XLog('ERR! '+PeerAddr+' > '+Request.Host+
           ' "'+Request.Method+' '+Request.URI+'"'+
           ' 0'+
           ' REF "'+Request.Referer+'"'+
           ' AGENT "'+Request.Agent+'" '+
           '> DISCONNECTED while receiving a Request ('+IntToStr(Request.DataIn)+' of '+IntToStr(Request.DataSize)+' bytes received).');
      end
    else if Response.DataSize>Response.DataOut then
      begin
      // did not send a complete result
      XLog('ERR! '+PeerAddr+' > '+Request.Host+
           ' "'+Request.Method+' '+Request.URI+'"'+
           ' -'+IntToStr(Response.DataSize-Response.DataOut)+
           ' REF "'+Request.Referer+'"'+
           ' AGENT "'+Request.Agent+'" '+
           '> DISCONNECTED while sending a Result ('+IntToStr(Response.DataOut)+' of '+IntToStr(Response.DataSize)+' bytes sent).');
      end;
    end;
  end;

procedure TPHP_Provider.DataModuleCreate(Sender: TObject);
  begin
  ExtList:=TStringList.Create;
  end;

procedure TPHP_Provider.DataModuleDestroy(Sender: TObject);
  begin
  PHP_Provider:=nil;

  ClearExts;
  ExtList.Free;
  end;

initialization
finalization
if assigned(PHP_Provider) then
  begin
  PHP_Provider.Free;
  PHP_Provider:=nil;
  end;
end.
