{##############################################################################}
{# NexusDB - RTC: nxRTCServer.pas %NXDBVERSION%                               #}
{# Copyright (c) Nexus Database Systems Pty. Ltd. %COPYRIGHT%                 #}
{# All rights reserved.                                                       #}
{##############################################################################}
{# NexusDB - RTC: Server                                                      #}
{##############################################################################}

unit nxRTCServer;

{$I nxDefine.inc}

interface

uses
  Classes,
  nxllMemoryManager,
  nxllTypes,
  nxllThread,
  nxllStreams,
  nxllSync,
  nxllTransport,

  rtcTransports,
  nxRTCCommon;

type
  EnxRTCPluginCommandHandlerException = class(EnxPluginCommandHandlerException);

  TnxRTCPluginCommandHandler = class(TnxBasePluginCommandHandler)
  protected {private}
    rpchServer                 : TComponent;
    procedure rpchSetServer(const aValue: TComponent);
  protected
    procedure scInitializing; override;

    procedure nxRTCTunnel(var Msg : TnxDataMessage);
  public
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    procedure bpchProcess(aMsg     : PnxDataMessage;
                      var aHandled : Boolean); override;

  published
    property RTCServer: TComponent
      read rpchServer
      write rpchSetServer;
  end;

implementation

uses
  Windows,
  nxllBde,
  SysUtils,
  nxllConvertException,
  nxllComponent;

resourcestring
  rsNoServerAssigend                        = 'No Server assigend';
  rsTheServerMustSupportIRTCMessageReceiver = 'The Server must support IRTCMessageReceiver';

{===TnxRTCPluginCommandHandler=======================================}
procedure TnxRTCPluginCommandHandler.bpchProcess(aMsg     : PnxDataMessage;
                                             var aHandled : Boolean);
begin
  aHandled := True;
  case aMsg^.dmMsg of
    nxnmRTCTunnel:
      nxRTCTunnel(aMsg^);
  else
    aHandled := False;
  end;
end;
{------------------------------------------------------------------------------}
procedure TnxRTCPluginCommandHandler.Notification(AComponent : TComponent;
                                                  Operation  : TOperation);
begin
  inherited;
  if (Operation = opRemove) and (aComponent = rpchServer) then
    RTCServer := nil;
end;
{------------------------------------------------------------------------------}
procedure TnxRTCPluginCommandHandler.nxRTCTunnel(var Msg: TnxDataMessage);
var
  Request : TnxDataMessageStream;
  Reply   : TnxMemoryStream;
  Error   : TnxResult;
begin
  Reply := TnxMemoryStream.Create;
  try

    if Assigned(rpchServer) then begin
      Request := TnxDataMessageStream.Create(Msg);
      try
        try
          (rpchServer as IRTCMessageReceiver).ProcessMessage(Request, Reply);
          Error := DBIERR_NONE;
        except
          on E: Exception do
            Error := ConvertException(E, lcEventLog);
        end;
      finally
        nxFreeAndNil(Request);
      end;
    end else
      Error := DBIERR_NOSERVERSW;

    if Error = DBIERR_NONE then
      TnxBaseTransport.Reply(nxnmRTCTunnel, Error, Reply.Memory, Reply.Size)
    else
      TnxBaseTransport.Reply(nxnmRTCTunnel, Error, nil, 0);

  finally
    nxFreeAndNil(Reply);
  end;
end;
{------------------------------------------------------------------------------}
procedure TnxRTCPluginCommandHandler.rpchSetServer(const aValue: TComponent);
begin
  scCheckInactive;

  if aValue <> rpchServer then begin
    if Assigned(rpchServer) then
      rpchServer.RemoveFreeNotification(Self);

    rpchServer := aValue;

    if Assigned(rpchServer) then
      rpchServer.FreeNotification(Self);
  end;
end;
{------------------------------------------------------------------------------}
procedure TnxRTCPluginCommandHandler.scInitializing;
{$IFNDEF DCC6OrLater}
var
  Dummy : IUnknown;
{$ENDIF}
begin
  inherited;

  if not Assigned(rpchServer) then
    raise EnxRTCPluginCommandHandlerException.nxcCreate(Self, rsNoServerAssigend);
  if not Supports(rpchServer, IRTCMessageReceiver{$IFNDEF DCC6OrLater}, Dummy{$ENDIF}) then
    raise EnxRTCPluginCommandHandlerException.nxcCreate(Self, rsTheServerMustSupportIRTCMessageReceiver);
end;
{==============================================================================}

end.
